"""Higher level wrapper to run training and parameter optimizations."""
import time
import warnings
from collections import defaultdict
from collections.abc import Iterator
from contextlib import AbstractContextManager, nullcontext
from functools import partial
from itertools import product
from tempfile import TemporaryDirectory
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import numpy as np
from joblib import Memory, Parallel
from numpy.ma import MaskedArray
from scipy.stats import rankdata
from sklearn.model_selection import BaseCrossValidator, ParameterGrid, check_cv
from tqdm.auto import tqdm
from typing_extensions import Self

from tpcp._algorithm_utils import (
    OPTIMIZE_METHOD_INDICATOR,
    _check_safe_optimize,
    _split_returns,
)
from tpcp._base import NOTHING, _get_annotated_fields_of_type
from tpcp._dataset import DatasetT
from tpcp._optimize import BaseOptimize
from tpcp._parameters import Parameter, _ParaTypes
from tpcp._pipeline import OptimizablePipelineT, PipelineT
from tpcp._utils._general import (
    _aggregate_final_results,
    _normalize_score_results,
    _passthrough,
    _prefix_para_dict,
    _split_hyper_and_pure_parameters,
)
from tpcp._utils._score import _optimize_and_score, _score
from tpcp.exceptions import PotentialUserErrorWarning
from tpcp.parallel import delayed
from tpcp.validate._scorer import ScorerTypes, _validate_scorer

if TYPE_CHECKING:
    from tpcp import OptimizablePipeline

T = TypeVar("T")


class DummyOptimize(BaseOptimize[PipelineT, DatasetT]):
    """Provide API compatibility for SimplePipelines in optimize wrappers.

    This is a simple dummy Optimizer that will **not** optimize anything, but just provide the correct API so that
    pipelines that do not have the possibility to be optimized can be passed to wrappers like
    :func:`tpcp.validate.cross_validate`.


    Parameters
    ----------
    pipeline
        The pipeline to wrap.
        It will not be optimized in any way, but simply copied to `self.optimized_pipeline_` if `optimize` is called.

    Other Parameters
    ----------------
    dataset
        The dataset used for optimization.
        As no optimization is performed, this will be ignored.

    Attributes
    ----------
    optimized_pipeline_
        The optimized version of the pipeline.
        In case of this class, this is just an unmodified clone of the input pipeline.

    """

    pipeline: Parameter[PipelineT]

    optimized_pipeline_: PipelineT

    def __init__(self, pipeline: PipelineT) -> None:
        self.pipeline = pipeline

    def optimize(self, dataset: DatasetT, **optimize_params: Any) -> Self:  # noqa: ARG002
        """Run the "dummy" optimization.

        Parameters
        ----------
        dataset
            The parameter is ignored, as no real optimization is performed
        optimize_params
            The parameter is ignored, as no real optimization is performed

        Returns
        -------
        self
            The class instance with all result attributes populated

        """
        self.dataset = dataset
        if hasattr(self.pipeline, "self_optimize"):
            warnings.warn(
                "You are using `DummyOptimize` with a pipeline that implements `self_optimize` and, hence, indicates "
                "that the pipeline can be optimized. "
                "`DummyOptimize` does never call this method and skips any optimization steps! "
                "Use `Optimize` if you actually want to optimize your pipeline.",
                PotentialUserErrorWarning,
                stacklevel=2,
            )
        self.optimized_pipeline_ = self.pipeline.clone()
        return self


class Optimize(BaseOptimize[OptimizablePipelineT, DatasetT]):
    """Run a generic self-optimization on the pipeline.

    This is a simple wrapper for pipelines that already implement a `self_optimize` method.
    This wrapper can be used to ensure that these algorithms can be optimized with the same interface as other
    optimization methods and can hence be used in methods like :func:`tpcp.validate.cross_validate`.

    Optimize will never modify the original pipeline, but will store a copy of the optimized pipeline as
    `optimized_pipeline_`.

    If `safe_optimize` is True, the wrapper applies the same runtime checks as provided by
    :func:`~tpcp.make_optimize_safe`.

    Parameters
    ----------
    pipeline
        The pipeline to optimize. The pipeline must implement `self_optimize` to optimize its own input parameters.
    safe_optimize
        If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
        implemented.
        See :func:`~tpcp.make_optimize_safe` for more info.
    optimize_with_info
        If True, Optimize will try to call `self_optimize_with_info` by default and will fall back to `self_optimize`.
        If you want to force optimize to use `self_optimize`, even if an implementation of `self_optimize_with_info`
        exists, set this parameter to False.

    Other Parameters
    ----------------
    dataset
        The dataset used for optimization.

    Attributes
    ----------
    optimized_pipeline_
        The optimized version of the pipeline.
        That is a copy of the input pipeline with modified params.
    optimization_info_
        If the optimized pipeline implements a `self_optimize_with_info` method, this parameter contains the
        additional information provided as second return value from this method.

    """

    pipeline: Parameter[OptimizablePipelineT]
    safe_optimize: bool
    optimize_with_info: bool

    optimized_pipeline_: OptimizablePipelineT
    optimization_info_: Any

    def __init__(
        self,
        pipeline: OptimizablePipelineT,
        *,
        safe_optimize: bool = True,
        optimize_with_info: bool = True,
    ) -> None:
        self.pipeline = pipeline
        self.safe_optimize = safe_optimize
        self.optimize_with_info = optimize_with_info

    def optimize(self, dataset: DatasetT, **optimize_params: Any) -> Self:
        """Run the self-optimization defined by the pipeline.

        The optimized version of the pipeline is stored as `self.optimized_pipeline_`.

        Parameters
        ----------
        dataset
            An instance of a :class:`~tpcp.Dataset` containing one or multiple data points that can
            be used for optimization.
            The structure of the data and the available reference information will depend on the dataset.
        optimize_params
            Additional parameter for the optimization process.
            They are forwarded to `pipeline.self_optimize`.

        Returns
        -------
        self
            The class instance with all result attributes populated

        """
        self.dataset = dataset
        if not hasattr(self.pipeline, "self_optimize"):
            raise ValueError(
                "To use `Optimize` with a pipeline, the pipeline needs to implement a `self_optimize` method."
            )
        # We clone just to make sure runs are independent
        pipeline: OptimizablePipeline = self.pipeline.clone()
        method = pipeline.self_optimize_with_info if self.optimize_with_info else pipeline.self_optimize
        if self.safe_optimize is True:
            # We check here, if the pipeline already has the safe decorator and if yes just call it.
            if getattr(method, OPTIMIZE_METHOD_INDICATOR, False) is True:
                optimized_pipeline, other_info = _split_returns(method(dataset, **optimize_params))
            else:
                optimized_pipeline, other_info = _split_returns(
                    _check_safe_optimize(pipeline, method, dataset, **optimize_params)
                )
        else:
            optimized_pipeline, other_info = _split_returns(method(dataset, **optimize_params))
        # We clone again, just to be sure
        self.optimized_pipeline_ = optimized_pipeline.clone()
        if self.optimize_with_info and other_info is not NOTHING:
            self.optimization_info_ = other_info
        return self


class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT, T]):
    """Perform a grid search over various parameters.

    This scores the pipeline for every combination of data points in the provided dataset and parameter combinations
    in the `parameter_grid`.
    The scores over the entire dataset are then aggregated for each parameter combination.
    By default, this aggregation is a simple average.

    .. note::
        This is different to how grid search works in many other cases:
        Usually, the performance parameter would be calculated on all data points at once.
        Here, each data point represents an entire participant or recording (depending on the dataset).
        Therefore, the pipeline and the scoring method are expected to provide a result/score per data point
        in the dataset.
        Note that it is still open to your interpretation what you consider a "data point" in the context of your
        analysis. The `run` method of the pipeline can still process multiple data points, e.g., gait tests, in a loop
        and generate a single output if you consider a single participant *one data point*.

    Parameters
    ----------
    pipeline
        The pipeline object to optimize
    parameter_grid
        A sklearn parameter grid to define the search space.
    scoring
        A callable that can score a single data point given a pipeline.
        This function should return either a single score or a dictionary of scores.
        If scoring is `None` the default `score` method of the pipeline is used instead.

        Note that if scoring returns a dictionary, `return_optimized` must be set to the name of the score that
        should be used for ranking.
    n_jobs
        The number of processes that should be used to parallelize the search.
        `None` means 1 while -1 means as many as logical processing cores.
    pre_dispatch
        The number of jobs that should be pre dispatched.
        For an explanation see the documentation of :class:`~sklearn.model_selection.GridSearchCV`
    return_optimized
        If True, a pipeline object with the overall best params is created and stored as `optimized_pipeline_`.
        If `scoring` returns a dictionary of score values, this must be a `str` corresponding to the name of the
        score that should be used to rank the results.
        If False, the respective result attributes will not be populated.
        If multiple parameter combinations have the same score, the one tested first will be used.
        By default, the value with the best `rank` (i.e. higher score) is used.
        If you want to select the value with the lowest score, set `return_optimized` to the name of the score prefixed
        with a minus sign, e.g. `-rmse`.
        In case of a single score, use `-score` to select the value with the lowest score.
    progress_bar
        True/False to enable/disable a tqdm progress bar.

    Other Parameters
    ----------------
    dataset
        The dataset instance passed to the optimize method

    Attributes
    ----------
    gs_results_
        A dictionary summarizing all results of the gridsearch.
        The format of this dictionary is designed to be directly passed into the :class:`~pandas.DataFrame` constructor.
        Each column then represents the result for one set of parameters

        The dictionary contains the following entries:

        param_*
            The value of a respective parameter
        params
            A dictionary representing all parameters
        score / {scorer-name}
            The aggregated value of a score over all data-points.
            If a single score is used for scoring, then the generic name "score" is used.
            Otherwise, multiple columns with the name of the respective scorer exist
        rank_score / rank_{scorer-name}
            A sorting for each score from the highest to the lowest value.
            If lower or higher values are better, depends on the scoring function and needs to be interpreted
            accordingly.
        single_score / single_{scorer-name}
            The individual scores per data point for each parameter combination.
            This is a list of values with the `len(dataset)`.
        data_labels
            A list of data labels in the order the single score values are provided.
            These can be used to associate the `single_score` values with a certain data point.
    optimized_pipeline_
        An instance of the input pipeline with the best parameter set.
        This is only available if `return_optimized` is not False.
    best_params_
        The parameter dict that resulted in the best result.
        This is only available if `return_optimized` is not False.
    best_index_
        The index of the result row in the output.
        This is only available if `return_optimized` is not False.
    best_score_
        The score of the best result.
        In a multimetric case, only the value of the scorer specified by `return_optimized` is provided.
        This is only available if `return_optimized` is not False.
    multimetric_
        If the scorer returned multiple scores

    """

    parameter_grid: ParameterGrid
    scoring: ScorerTypes[PipelineT, DatasetT, T]
    n_jobs: Optional[int]
    return_optimized: Union[bool, str]
    pre_dispatch: Union[int, str]
    progress_bar: bool

    gs_results_: dict[str, Any]
    best_params_: dict[str, Any]
    best_index_: int
    best_score_: float
    multimetric_: bool

    def __init__(
        self,
        pipeline: PipelineT,
        parameter_grid: ParameterGrid,
        *,
        scoring: ScorerTypes[PipelineT, DatasetT, T] = None,
        n_jobs: Optional[int] = None,
        return_optimized: Union[bool, str] = True,
        pre_dispatch: Union[int, str] = "n_jobs",
        progress_bar: bool = True,
    ) -> None:
        self.pipeline = pipeline
        self.parameter_grid = parameter_grid
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.pre_dispatch = pre_dispatch
        self.return_optimized = return_optimized
        self.progress_bar = progress_bar

    def optimize(self, dataset: DatasetT, **_: Any) -> Self:
        """Run the grid search over the dataset and find the best parameter combination.

        Parameters
        ----------
        dataset
            The dataset used for optimization.

        """
        self.dataset = dataset
        scoring = _validate_scorer(self.scoring, self.pipeline)

        # We use a similar structure as sklearn's GridSearchCV here, but instead of calling something equivalent to
        # `fit_score`, we call `score`, which just applies and scores the pipeline on the entirety of our dataset as
        # we do not need a "train" step.
        # Our main loop just loops over all parameter combinations and the `_score` function then applies the parameter
        # combination to the pipeline and scores the resulting pipeline on the dataset, by passing the entire dataset
        # and the pipeline to the scorer.
        # Looping over the individual data points in the dataset and aggregating the scores is handled by the scorer
        # itself.
        # If not explicitly changed the scorer is an instance of `Scorer` that wraps the actual `scoring`
        # function provided by the user.
        if self.progress_bar:
            pbar = partial(tqdm, total=len(self.parameter_grid), desc="Parameter Combinations")
        else:
            pbar = _passthrough

        parallel = Parallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch, return_as="generator")
        with parallel:
            # Evaluate each parameter combination
            results = list(
                pbar(
                    parallel(
                        delayed(_score)(
                            self.pipeline.clone(),
                            dataset,
                            scoring,
                            paras,
                            return_parameters=True,
                            return_data_labels=True,
                            return_times=True,
                            error_info=f"This error occurred for the following parameter:\n\n{paras}",
                        )
                        for paras in self.parameter_grid
                    )
                )
            )
        assert results is not None  # For the typechecker
        # We check here if all results are dicts. We only check the dtype of the first value, as the scorer should
        # have handled issues with non-uniform cases already.
        first_test_score = results[0]["scores"]
        self.multimetric_ = isinstance(first_test_score, dict)

        results = self._format_results(
            list(self.parameter_grid),
            results,
        )

        reverse_ranking, return_optimized = _validate_return_optimized(
            self.return_optimized, self.multimetric_, first_test_score
        )
        if return_optimized:
            (
                self.best_index_,
                self.best_score_,
                self.best_params_,
            ) = _extract_return_optimize_info(
                return_optimized,
                reverse_ranking,
                results,
                rank_prefix="rank_",
                score_prefix="",
            )
            # We clone twice, in case one of the params was itself an algorithm.
            self.optimized_pipeline_ = self.pipeline.clone().set_params(**self.best_params_).clone()

        self.gs_results_ = results

        return self

    def _format_results(self, candidate_params, out):
        """Format the final result dict.

        This function is adapted based on sklearn's `BaseSearchCV`
        """
        n_candidates = len(candidate_params)
        out = _aggregate_final_results(out)

        results = {}

        scores_dict = _normalize_score_results(out["scores"]) or {}
        single_scores_dict = _normalize_score_results(out["single_scores"]) or {}
        for c, v in scores_dict.items():
            results[c] = v
            results[f"rank_{c}"] = np.asarray(rankdata(-v, method="min"), dtype=np.int32)
        for c, _v in single_scores_dict.items():
            # Because of custom aggregators, it can be that single scores dicts have different keys than the aggregated
            # scores.
            results[f"single_{c}"] = single_scores_dict[c]

        results["data_labels"] = out["data_labels"]
        results["score_time"] = out["score_time"]

        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(
            partial(
                MaskedArray,
                np.empty(
                    n_candidates,
                ),
                mask=True,
                dtype=object,
            )
        )
        for cand_idx, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurrence of `name`.
                # Setting the value at an index also unmasks that index
                param_results[f"param_{name}"][cand_idx] = value

        results.update(param_results)
        # Store a list of param dicts at the key 'params'
        results["params"] = candidate_params

        return results


class GridSearchCV(
    BaseOptimize[OptimizablePipelineT, DatasetT],
    Generic[OptimizablePipelineT, DatasetT, T],
):
    """Exhaustive (hyper)parameter search using a cross validation based score to optimize pipeline parameters.

    This class follows as much as possible the interface of :func:`~sklearn.model_selection.GridSearchCV`.
    If the `tpcp` documentation is missing some information, the respective documentation of `sklearn` might be helpful.

    Compared to the `sklearn` implementation this method uses a couple of `tpcp`-specific optimizations and
    quality-of-life improvements.

    Parameters
    ----------
    pipeline
        A tpcp pipeline implementing `self_optimize`.
    parameter_grid
        A sklearn parameter grid to define the search space for the grid search.
    scoring
        A callable that can score a single data point given a pipeline.
        This function should return either a single score or a dictionary of scores.
        If scoring is `None` the default `score` method of the pipeline is used instead.

        .. note:: If scoring returns a dictionary, `return_optimized` must be set to the name of the score that
                  should be used for ranking.
    return_optimized
        If True, a pipeline object with the overall best parameters is created and re-optimized using all provided data
        as input.
        The optimized pipeline object is stored as `optimized_pipeline_`.
        If `scoring` returns a dictionary of score values, this must be a str corresponding to the name of the
        score that should be used to rank the results.
        If False, the respective result attributes will not be populated.
        If multiple parameter combinations have the same mean score over all CV folds, the one tested first will be
        used.
        By default, the value with the best `rank` (i.e. higher score) is used.
        If you want to select the value with the lowest score, set `return_optimized` to the name of the score prefixed
        with a minus sign, e.g. `-rmse`.
        In case of a single score, use `-score` to select the value with the lowest score.
    cv
        An integer specifying the number of folds in a K-Fold cross validation or a valid cross validation helper.
        The default (`None`) will result in a 5-fold cross validation.
        For further inputs check the `sklearn`
        `documentation <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html>`_.
    pure_parameters
        .. warning::
            Do not use this option unless you fully understand it!

        A list of parameter names (named in the `parameter_grid`) that do not affect training aka are not
        hyperparameters.
        This information can be used for massive performance improvements, as the training does not need to be
        repeated if one of these parameters changes.
        However, setting it incorrectly can lead detect errors that are very hard to detect in the final results.

        Instead of passing a list of names, you can also just set the value to `True`.
        In this case all parameters of the provided pipeline that are marked as :class:`~tpcp.PureParameter` are used.
        Note that pure parameters of nested objects are not considered, but only top-level attributes.
        If you need to mark nested parameters as pure, use the first method and pass the names (with `__`) as part of
        the list of names.

        For more information on this approach see the :func:`evaluation guide <algorithm_evaluation>`.
    return_train_score
        If True the performance on the train score is returned in addition to the test score performance.
        Note, that this increases the runtime.
        If `True`, the fields `train_score`, and `train_score_single` are available in the results.
    verbose
        Control the verbosity of information printed during the optimization (larger number -> higher verbosity).
        At the moment this will only affect the caching done, when `pure_parameter_names` are provided.
    n_jobs
        The number of parallel jobs.
        The default (`None`) means 1 job at the time, hence, no parallel computing.
        -1 means as many as logical processing cores.
        One job is created per cv + para combi combination.
    pre_dispatch
        The number of jobs that should be pre dispatched.
        For an explanation see the documentation of :class:`~sklearn.model_selection.GridSearchCV`
    progress_bar
        True/False to enable/disable a tqdm progressbar.
    safe_optimize
        If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
        implemented.
        See :func:`~tpcp.make_optimize_safe` for more info.
    optimize_with_info
        If True, Optimize will try to call `self_optimize_with_info` by default and will fall back to `self_optimize`.
        If you want to force the optimization to use `self_optimize`, even if an implementation of
        `self_optimize_with_info` exists, set this parameter to False.

    Other Parameters
    ----------------
    dataset
        The dataset instance passed to the optimize method
    groups
        The groups passed to the optimize method
    mock_labels
        The mock labels passed to the optimize method

    Attributes
    ----------
    cv_results_
        A dictionary summarizing all results of the gridsearch.
        The format of this dictionary is designed to be directly passed into the :class:`~pandas.DataFrame` constructor.
        Each column then represents the result for one set of parameters.

        The dictionary contains the following entries:

        param_{parameter_name}
            The value of a respective parameter.
        params
            A dictionary representing all parameters.
        mean_test_score / mean_test_{scorer_name}
            The average test score over all folds.
            If a single score is used for scoring, then the generic name "score" is used.
            Otherwise, multiple columns with the name of the respective scorer exist.
        std_test_score / std_test_{scorer_name}
            The std of the test scores over all folds.
        rank_test_score / rank_{scorer_name}
            The rank of the mean test score assuming higher values are better.
        split{n}_test_score / split{n}_test_{scorer_name}
            The performance on the test set in fold n.
        split{n}_test_single_score / split{n}_test_single_{scorer_name}
            The performance in fold n on every single data point in the test set.
        split{n}_test_data_labels
            The ids of the data points used in the test set of fold n.
        mean_train_score / mean_train_{scorer_name}
            The average train score over all folds.
        std_train_score / std_train_{scorer_name}
            The std of the train scores over all folds.
        split{n}_train_score / split{n}_train_{scorer_name}
            The performance on the train set in fold n.
        rank_train_score / rank_{scorer_name}
            The rank of the mean train score assuming higher values are better.
        split{n}_train_single_score / split{n}_train_single_{scorer_name}
            The performance in fold n on every single datapoint in the train set.
        split{n}_train_data_labels
            The ids of the data points used in the train set of fold n.
        mean_{optimize/score}_time
            Average time over all folds spent for optimization and scoring, respectively.
        std_{optimize/score}_time
            Standard deviation of the optimize/score times over all folds.

    optimized_pipeline_
        An instance of the input pipeline with the best parameter set.
        This is only available if `return_optimized` is not False.
    best_params_
        The parameter dict that resulted in the best result.
        This is only available if `return_optimized` is not False.
    best_index_
        The index of the result row in the output.
        This is only available if `return_optimized` is not False.
    best_score_
        The score of the best result.
        In a multimetric case, only the value of the scorer specified by `return_optimized` is provided.
        This is only available if `return_optimized` is not False.
    multimetric_
        If the scorer returned multiple scores
    final_optimize_time_
        Time spent to perform the final optimization on all data.
        This is only available if `return_optimized` is not False.

    """

    pipeline: OptimizablePipelineT
    parameter_grid: ParameterGrid
    scoring: ScorerTypes[OptimizablePipelineT, DatasetT, T]
    return_optimized: Union[bool, str]
    cv: Optional[Union[int, BaseCrossValidator, Iterator]]
    pure_parameters: Union[bool, list[str]]
    return_train_score: bool
    verbose: int
    n_jobs: Optional[int]
    pre_dispatch: Union[int, str]
    progress_bar: bool
    safe_optimize: bool
    optimize_with_info: bool

    groups: Optional[list[Union[str, tuple[str, ...]]]]
    mock_labels: Optional[list[Union[str, tuple[str, ...]]]]

    cv_results_: dict[str, Any]
    best_params_: dict[str, Any]
    best_index_: int
    best_score_: float
    multimetric_: bool
    final_optimize_time_: float

    def __init__(
        self,
        pipeline: OptimizablePipelineT,
        parameter_grid: ParameterGrid,
        *,
        scoring: ScorerTypes[OptimizablePipelineT, DatasetT, T] = None,
        return_optimized: Union[bool, str] = True,
        cv: Optional[Union[int, BaseCrossValidator, Iterator]] = None,
        pure_parameters: Union[bool, list[str]] = False,
        return_train_score: bool = False,
        verbose: int = 0,
        n_jobs: Optional[int] = None,
        pre_dispatch: Union[int, str] = "n_jobs",
        progress_bar: bool = True,
        safe_optimize: bool = True,
        optimize_with_info: bool = True,
    ) -> None:
        self.pipeline = pipeline
        self.parameter_grid = parameter_grid
        self.scoring = scoring
        self.return_optimized = return_optimized
        self.cv = cv
        self.pure_parameters = pure_parameters
        self.return_train_score = return_train_score
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.pre_dispatch = pre_dispatch
        self.progress_bar = progress_bar
        self.safe_optimize = safe_optimize
        self.optimize_with_info = optimize_with_info

    def optimize(self, dataset: DatasetT, *, groups=None, mock_labels=None, **optimize_params) -> Self:
        """Run the GridSearchCV on the given dataset.

        Parameters
        ----------
        dataset
            The dataset to optimize on.
        groups
            An optional set of group labels that are passed to the cross-validation helper.
        mock_labels
            An optional set of mocked labels that are passed to the cross-validation helper as the `y` parameter.
            This can be helpful in combination with the `Stratified*Fold` cross-validation helpers, that use the `y`
            parameter to stratify the folds.

        """
        self.dataset = dataset
        self.groups = groups
        self.mock_labels = mock_labels

        scoring = _validate_scorer(self.scoring, self.pipeline)

        cv_checked: BaseCrossValidator = check_cv(self.cv, None, classifier=True)
        n_splits = cv_checked.get_n_splits(dataset, mock_labels, groups=groups)

        # We need to wrap our pipeline for a consistent interface.
        # In the future we might be able to allow objects with optimizer Interface as input directly.
        optimizer = Optimize(
            self.pipeline,
            safe_optimize=self.safe_optimize,
            optimize_with_info=self.optimize_with_info,
        )

        # For each para combi, we separate the pure parameters (parameters that do not affect the optimization) and
        # the hyperparameters.
        # This allows for massive caching optimizations in the `_optimize_and_score`.
        pure_parameters: Optional[list[str]]
        if self.pure_parameters is False:
            pure_parameters = None
        elif self.pure_parameters is True:
            pure_parameters = _get_annotated_fields_of_type(self.pipeline, _ParaTypes.PURE)
        elif isinstance(self.pure_parameters, list):
            pure_parameters = self.pure_parameters
        else:
            raise ValueError(
                "`self.pure_parameters` must either be a List of field names (nested are allowed) or True/False."
            )

        parameters = list(self.parameter_grid)
        split_parameters = _split_hyper_and_pure_parameters(parameters, pure_parameters)
        parameter_prefix = "pipeline__"
        combinations = list(
            product(
                enumerate(split_parameters),
                enumerate(cv_checked.split(dataset, mock_labels, groups=groups)),
            )
        )

        pbar = partial(tqdm, total=len(combinations), desc="Split-Para Combos") if self.progress_bar else _passthrough

        # To enable the pure parameter performance improvement, we need to create a joblib cache in a temp dir that
        # is deleted after the run.
        # We only allow a temporary cache here, because the method that is cached internally is generic and the cache
        # might not be correctly invalidated, if GridSearchCv is called with a different pipeline or when the
        # pipeline itself is modified.
        tmp_dir_context: Union[AbstractContextManager[None], TemporaryDirectory] = nullcontext()
        if pure_parameters:
            tmp_dir_context = TemporaryDirectory("joblib_tpcp_cache")
        with tmp_dir_context as cachedir:
            tmp_cache = Memory(cachedir, verbose=self.verbose) if cachedir else None
            parallel = Parallel(
                n_jobs=self.n_jobs,
                pre_dispatch=self.pre_dispatch,
                return_as="generator",
            )
            # We use a similar structure to sklearn's GridSearchCV here (see GridSearch for more info).
            with parallel:
                # Evaluate each parameter combination
                out = list(
                    pbar(
                        parallel(
                            delayed(_optimize_and_score)(
                                optimizer.clone(),
                                scoring,
                                dataset[train],
                                dataset[test],
                                optimize_params=optimize_params,
                                hyperparameters=_prefix_para_dict(hyper_paras, parameter_prefix),
                                pure_parameters=_prefix_para_dict(pure_paras, parameter_prefix),
                                return_train_score=self.return_train_score,
                                return_parameters=False,
                                return_data_labels=True,
                                return_times=True,
                                memory=tmp_cache,
                                error_info=f"This error occurred in fold {split_idx} with parameters candidate "
                                f"{cand_idx}.",
                            )
                            for (cand_idx, (hyper_paras, pure_paras)), (
                                split_idx,
                                (train, test),
                            ) in combinations
                        )
                    )
                )
        assert out is not None  # For the type checker
        results = self._format_results(parameters, n_splits, out)
        self.cv_results_ = results

        first_test_score = out[0]["test_scores"]
        self.multimetric_ = isinstance(first_test_score, dict)
        reverse_ranking, return_optimized = _validate_return_optimized(
            self.return_optimized, self.multimetric_, first_test_score
        )
        if return_optimized:
            (
                self.best_index_,
                self.best_score_,
                self.best_params_,
            ) = _extract_return_optimize_info(
                return_optimized,
                reverse_ranking,
                results,
                rank_prefix="rank_test_",
                score_prefix="mean_test_",
            )
            # We clone twice, in case one of the params was itself an algorithm.
            best_optimizer = Optimize(
                self.pipeline.clone().set_params(**self.best_params_).clone(),
                safe_optimize=self.safe_optimize,
            )
            final_optimize_start_time = time.time()
            optimize_params_clean = optimize_params or {}
            self.optimized_pipeline_ = best_optimizer.optimize(dataset, **optimize_params_clean).optimized_pipeline_
            self.final_optimize_time_ = final_optimize_start_time - time.time()

        return self

    def _format_results(self, candidate_params, n_splits, out, more_results=None):  # noqa: C901
        """Format the final result dict.

        This function is adapted based on sklearn's `BaseSearchCV`.
        """
        n_candidates = len(candidate_params)
        out = _aggregate_final_results(out)

        results = dict(more_results or {})

        def _store_non_numeric(key_name: str, array):
            """Store non-numeric scores/times to the cv_results_."""
            # We avoid performing any sort of conversion into numpy arrays as this might modify the dtypes.
            # Instead we use list comprehension to do the reshaping.
            # The result is a list of lists and not a numpy array.
            #
            # Note that the results were produced by iterating first by splits and then by parameters.
            # We do the same here, but directly transpose the results, as we need to access the data per parameters
            # afterwards.
            iterable_array = iter(array)
            array = [[next(iterable_array) for _ in range(n_splits)] for _ in range(n_candidates)]
            # "Transpose" the array
            array = map(list, zip(*array))

            for split_idx, split in enumerate(array):
                # Uses closure to alter the results
                results[f"split{split_idx}_{key_name}"] = split

        def _store(key_name: str, array, weights=None, splits=False, rank=False):
            """Store numeric scores/times to the cv_results_."""
            # When iterated first by splits, then by parameters
            # We want `array` to have `n_candidates` rows and `n_splits` cols.
            array = np.array(array, dtype=np.float64).reshape(n_candidates, n_splits)

            if splits:
                for split_idx in range(n_splits):
                    # Uses closure to alter the results
                    results[f"split{split_idx}_{key_name}"] = array[:, split_idx]
            array_means = np.average(array, axis=1, weights=weights)
            results[f"mean_{key_name}"] = array_means

            if key_name.startswith(("train_", "test_")) and np.any(~np.isfinite(array_means)):
                warnings.warn(
                    f"One or more of the {key_name.split('_')[0]} scores are non-finite: {array_means}",
                    category=UserWarning,
                    stacklevel=2,
                )
            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(np.average((array - array_means[:, np.newaxis]) ** 2, axis=1, weights=weights))
            results[f"std_{key_name}"] = array_stds

            if rank:
                results[f"rank_{key_name}"] = np.asarray(rankdata(-array_means, method="min"), dtype=np.int32)

        _store("optimize_time", out["optimize_time"])
        _store("score_time", out["score_time"])
        _store_non_numeric("test_data_labels", out["test_data_labels"])
        _store_non_numeric("train_data_labels", out["train_data_labels"])
        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results: dict = defaultdict(
            partial(
                MaskedArray,
                np.empty(
                    n_candidates,
                ),
                mask=True,
                dtype=object,
            )
        )
        for cand_idx, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_{name}"` at the first occurrence of `name`.
                # Setting the value at an index also unmasks that index
                param_results[f"param_{name}"][cand_idx] = value

        results.update(param_results)
        # Store a list of param dicts at the key 'params'
        results["params"] = candidate_params

        test_scores_dict = _normalize_score_results(out["test_scores"]) or {}
        test_single_scores_dict = _normalize_score_results(out["test_single_scores"]) or {}

        for scorer_name in test_scores_dict:
            # Computed the (weighted) mean and std for test scores alone
            _store(
                f"test_{scorer_name}",
                test_scores_dict[scorer_name],
                splits=True,
                rank=True,
                weights=None,
            )
        for scorer_name in test_single_scores_dict:
            # Because of custom aggregators, it can be that single scores dicts have different keys than the aggregated
            # scores.
            _store_non_numeric(f"test_single_{scorer_name}", test_single_scores_dict[scorer_name])

        if self.return_train_score:
            train_scores_dict = _normalize_score_results(out["train_scores"])
            train_single_scores_dict = _normalize_score_results(out["train_single_scores"])
            for scorer_name in train_scores_dict:
                _store(f"train_{scorer_name}", train_scores_dict[scorer_name], splits=True)
            for scorer_name in train_single_scores_dict:
                _store_non_numeric(f"train_single_{scorer_name}", train_single_scores_dict[scorer_name])

        return results


def _validate_return_optimized(return_optimized, multi_metric, results) -> tuple[bool, Union[str, Literal[False]]]:
    """Check if `return_optimize` fits to the multimetric output of the scorer."""
    if return_optimized is False:
        return False, False
    # Sanitize return optimized
    reverse = False
    if isinstance(return_optimized, str) and return_optimized.startswith("-"):
        return_optimized = return_optimized[1:].strip()
        reverse = True
    if multi_metric is True:
        # In a multimetric case, return_optimized must either be False or a string
        if not isinstance(return_optimized, str) or return_optimized not in results:
            raise ValueError(
                "If multi-metric scoring is used, `return_optimized` must be a str specifying the score that "
                "should be used to select the best result."
            )
        return reverse, return_optimized
    # Single metric:
    if return_optimized == "score" or return_optimized is True:
        return reverse, "score"
    if isinstance(return_optimized, str):
        warnings.warn(
            "You set `return_optimized` to the name of a scorer, but the provided scorer only produces a "
            "single score."
            "`return_optimized` is set to True. "
            "The only allowed string value for `return_optimized` in a single metric case is `-score`, "
            "to invert the metric before score selection.",
            stacklevel=2,
        )
        return reverse, "score"
    raise ValueError("`return_optimized` must be a bool or explicitly `score` or `-score` in a single metric case.")


def _extract_return_optimize_info(
    return_optimized: str,
    reverse,
    results,
    rank_prefix: str = "rank_",
    score_prefix: str = "",
) -> tuple[int, float, dict[str, Any]]:
    """Extract the information from `return_optimized` and check if it is valid."""
    if reverse:
        best_index = (results[f"{rank_prefix}{return_optimized}"]).argmax()
    else:
        best_index = results[f"{rank_prefix}{return_optimized}"].argmin()
    best_score = results[f"{score_prefix}{return_optimized}"][best_index]
    best_params = results["params"][best_index]
    return best_index, best_score, best_params



"""Implementation of methods and classes to wrap the optimization Framework `Optuna`."""

try:
    import optuna
except ImportError as e:
    raise ImportError(
        "To use the tpcp Optuna interface, you first need to install optuna (`pip install optuna`)"
    ) from e

import multiprocessing
import warnings
from ast import literal_eval
from collections import defaultdict
from collections.abc import Sequence
from typing import Any, Callable, Optional, TypedDict, TypeVar, Union

import numpy as np
from joblib import Parallel
from optuna import Study
from optuna.study import StudyDirection
from optuna.study.study import ObjectiveFuncType
from optuna.trial import FrozenTrial, Trial
from typing_extensions import Self

from tpcp import OptimizablePipeline, clone
from tpcp._dataset import DatasetT
from tpcp._optimize import BaseOptimize
from tpcp._pipeline import PipelineT
from tpcp.optimize import Optimize
from tpcp.parallel import delayed
from tpcp.validate._scorer import ScorerTypes, _validate_scorer

__all__ = ["CustomOptunaOptimize", "CustomOptunaOptimizeT", "OptunaSearch"]

if multiprocessing.parent_process() is None:
    # We want to avoid spamming the user with warnings if they are running multiple processes
    warnings.warn(
        "The Optuna interface in tpcp is still experimental and we are testing if the workflow makes "
        "sense even for larger projects. "
        "This means, the interface for `CustomOptunaOptimize` will likely change in the future.",
        UserWarning,
        stacklevel=2,
    )


CustomOptunaOptimizeT = TypeVar("CustomOptunaOptimizeT", bound="_CustomOptunaOptimize")


def _split_trials(n_trials, n_jobs):
    n_per_job, remaining = divmod(n_trials, n_jobs)
    for _ in range(n_jobs):
        yield n_per_job + (1 if remaining > 0 else 0)
        remaining -= 1


class StudyParamsDict(TypedDict, total=False):
    """Type of the dictionary returned by `get_study_params`."""

    study_name: Optional[str]
    storage: Union[str, optuna.storages.BaseStorage]
    sampler: Optional[optuna.samplers.BaseSampler]
    pruner: Optional[optuna.pruners.BasePruner]
    direction: Optional[Union[str, StudyDirection]]
    load_if_exists: bool
    directions: Optional[Sequence[Union[str, StudyDirection]]]


class _CustomOptunaOptimize(BaseOptimize[PipelineT, DatasetT]):
    pipeline: PipelineT
    get_study_params: Callable[[int], StudyParamsDict]
    return_optimized: bool

    # Optuna Parameters that are directly forwarded to study.optimize
    n_trials: Optional[int]
    timeout: Optional[float]
    callbacks: Optional[list[Callable[[Study, FrozenTrial], None]]]
    gc_after_trial: bool
    show_progress_bar: bool
    n_jobs: int
    random_seed: Optional[int]

    eval_str_paras: Sequence[str]

    optimized_pipeline_: PipelineT
    random_seed_: int
    study_: Study

    @property
    def search_results_(self) -> dict[str, Sequence[Any]]:
        r"""Detailed results of the study.

        This basically contains the same information as `self.study_.trials_dataframe()`, with some small modifications:

        - columns starting with "params\_" are renamed to "param\_"
        - a new column called "params" containing all parameters as dict is added
        - "value" is renamed to score"
        - the score of pruned trials is set to `np.nan`

        These changes are made to make the output comparable to the output of :class:`~tpcp.optimize.GridSearch` and
        :class:`~tpcp.optimize.GridSearchCV`.
        """

        def rename_param_columns(name: str):
            param_start = "params_"
            if name.startswith(param_start):
                return "param_" + name[len(param_start) :]
            return name

        base_df = (
            self.study_.trials_dataframe()
            .drop("number", axis=1)
            .rename(columns={"value": "score"})
            .rename(columns=rename_param_columns)
        )
        base_df["params"] = [self.sanitize_params(t.params) for t in self.study_.trials]

        # If a trial is pruned we set the score to nan.
        # This is clearer than showing the last step value.
        # If this is required people should report them as user values in their objective function
        base_df.loc[base_df["state"] == "PRUNED", "score"] = np.nan

        return base_df.to_dict(orient="list")

    @property
    def best_params_(self) -> dict[str, Any]:
        """Parameters of the best trial in the :class:`~optuna.study.Study`."""
        return self.sanitize_params(self.study_.best_params)

    @property
    def best_score_(self) -> float:
        """Best score reached in the study."""
        return self.study_.best_value

    @property
    def best_trial_(self) -> FrozenTrial:
        """Best trial in the :class:`~optuna.study.Study`."""
        return self.study_.best_trial

    def optimize(self, dataset: DatasetT, **_: Any) -> Self:
        """Optimize the objective over the dataset and find the best parameter combination.

        This method calls `self.create_objective` to obtain the objective function that should be optimized.

        Parameters
        ----------
        dataset
            The dataset used for optimization.

        """
        if self.timeout is None and self.n_trials is None:
            raise ValueError(
                "You need to set either `timeout` or `n_trials` to a proper value."
                "Otherwise the optimization will not stop und run until infinity."
            )

        self.dataset = dataset

        objective = self._create_objective(self.pipeline, dataset=dataset)
        rng = np.random.default_rng()
        self.random_seed_ = self.random_seed if self.random_seed is not None else rng.integers(0, 100)

        study_params = self.get_study_params(self.random_seed_)
        self.study_ = optuna.create_study(**study_params)

        if not isinstance(self.study_._storage, optuna.storages.InMemoryStorage):
            warnings.warn(
                "You are using a persistent storage for the study. "
                "This means, that the study will be saved and likely be available after this script "
                "terminates. "
                "To avoid issues and make sure that runs of your optimizations are reproducible and "
                "independent, make sure to cleanup the persistent storage once the results are not needed "
                "anymore.\n"
                "You can use use `optuna.delete_study(study_name=opti_instance.study_.study_name, "
                "storage=opti_instance.study_._storage)`. "
                "Note that all result object that depend on the study are not available anymore after deletion.",
                stacklevel=2,
            )

        if self.n_jobs == 1:
            self._call_optimize(self.study_, objective)
        else:
            if isinstance(self.study_._storage, optuna.storages.InMemoryStorage):
                raise ValueError(
                    "You are using the InMemoryStorage with n_jobs > 1. "
                    "This will lead to problems as the storage is not persistent and thread safe. "
                    "Use a persistent database based storage instead."
                )
            if self.timeout is not None:
                raise ValueError(
                    "You are using timeout with n_jobs > 1. "
                    "This parameter does not make sense in this case. "
                    "You `n_trials` instead to limit for how long the optimization should run."
                )

            if self.show_progress_bar:
                warnings.warn(
                    "You are using a progress bar with n_jobs > 1. "
                    "This might lead to strange behaviour, as each process will launch its own process bar with "
                    "n_trials/n_jobs steps.",
                    stacklevel=2,
                )

            # This solution is based on the solution proposed here:
            # https://github.com/optuna/optuna/issues/2862
            def _multi_process_call_optimize(n_trials: int, seed: int):
                study_params_per_process = self.get_study_params(seed)
                if (name := study_params_per_process.get("study_name", None)) is not None:
                    assert (
                        name == self.study_.study_name
                    ), "Study name must be the same on every call of `self.get_study_params`"
                study = optuna.load_study(
                    # Name and storage come from the existing study, as things will break, if they are not consistent
                    # across processes
                    study_name=self.study_.study_name,
                    storage=self.study_._storage,
                    # Pruner and sampler come from the params, because they have updated random seeds
                    sampler=study_params_per_process.get("sampler", None),
                    pruner=study_params_per_process.get("pruner", None),
                )
                self._call_optimize_multi_process(study, objective, n_trials)

            # We call the objective function once in the main thread to make sure that the study is created correctly
            _multi_process_call_optimize(1, self.random_seed_)
            parallel = Parallel(self.n_jobs)
            parallel(
                delayed(_multi_process_call_optimize)(n_trials=n_trials_i, seed=self.random_seed_ + i + 1)
                for (i, n_trials_i) in enumerate(_split_trials(self.n_trials - 1, self.n_jobs))
            )

        if self.return_optimized:
            self.optimized_pipeline_ = self.return_optimized_pipeline(clone(self.pipeline), dataset, self.study_)
        return self

    def _call_optimize(self, study: Study, objective: ObjectiveFuncType):
        """Call the optuna study.

        This is a separate method to make it easy to modify how the study is called.
        """
        study.optimize(
            objective,
            n_trials=self.n_trials,
            timeout=self.timeout,
            callbacks=self.callbacks,
            gc_after_trial=self.gc_after_trial,
            show_progress_bar=self.show_progress_bar,
        )
        return study

    def _call_optimize_multi_process(self, study: Study, objective: ObjectiveFuncType, n_trials: int):
        study.optimize(
            objective,
            n_trials=n_trials,
            timeout=None,  # This does not work with multiprocessing
            callbacks=self.callbacks,
            gc_after_trial=self.gc_after_trial,
            show_progress_bar=self.show_progress_bar,  # This is a little strange. We warn about it above
        )
        return study

    def create_objective(self) -> Callable[[Trial, PipelineT, DatasetT], Union[float, Sequence[float]]]:
        """Return the objective function that should be optimized.

        This method should be implemented by a child class and return an objective function that is compatible with
        Optuna.
        However, compared to a normal Optuna objective function, the function should expect a pipeline and a dataset
        object as additional inputs to the optimization Trial object.
        """
        raise NotImplementedError()

    def sanitize_params(self, params: dict[str, Any]) -> dict[str, Any]:
        """Sanatize the parameters of a trial.

        This will apply the str evaluation controlled by `self.eval_str_paras` to the parameters.
        Call this method before passing the parameters to the pipeline in your objective function.
        """
        final_params = {}
        for k, v in params.items():
            if k in self.eval_str_paras:
                final_params[k] = literal_eval(v)
                continue
            final_params[k] = v
        return final_params

    def _create_objective(self, pipeline: PipelineT, dataset: DatasetT) -> ObjectiveFuncType:
        inner_objective = self.create_objective()

        def objective(trial: Trial):
            inner_pipe = clone(pipeline)
            return inner_objective(trial, inner_pipe, dataset)

        return objective

    def return_optimized_pipeline(self, pipeline: PipelineT, dataset: DatasetT, study: Study) -> PipelineT:
        """Return the pipeline with the best parameters of a study.

        This either just returns the pipeline with the best parameters set, or if the pipeline is a subclass of
        `OptimizablePipeline` it attempts a re-optimization of the pipeline using the provided dataset.

        This functionality is a sensible default, but it is expected to overwrite this method in custom subclasses,
        if specific behaviour is needed.

        Don't call this function on its own! It is only expected to be called internally by optimize.
        """
        # Pipeline that will be passed here is already cloned, so no need to clone again.
        pipeline_with_best_params = pipeline.set_params(**self.sanitize_params(study.best_params))
        if isinstance(pipeline_with_best_params, OptimizablePipeline):
            return Optimize(pipeline_with_best_params).optimize(dataset).optimized_pipeline_
        return pipeline_with_best_params


class CustomOptunaOptimize(_CustomOptunaOptimize[PipelineT, DatasetT]):
    """Base class for custom Optuna optimizer.

    This provides a relatively simple tpcp compatible interface to Optuna.
    You basically need to subclass this class and implement the `create_objective` method to return the objective
    function you want to optimize.
    The only difference to a normal objective function in Optuna is, that your objective here should expect a
    pipeline and a dataset object as second and third argument (see Example).
    If there are parameters you want to make customizable (e.g. which metric to optimize for), expose them in the
    `__init__` of your subclass.

    Depending on your usecase, your custom optimizers can be single use with a bunch of "hard-coded" logic, or you can
    try to make them more general, by exposing certain configurability.

    Parameters
    ----------
    pipeline
        A tpcp pipeline with some hyper-parameters that should be optimized.
        This can either be a normal pipeline or an optimizable-pipeline.
        This fully depends on your implementation of the `create_objective` method.
    get_study_params
        A callable that returns a dictionary with the parameters that should be used to create the study (i.e. passed
        to `optuna.create_study`).
        Creating the study is handled via a callable, instead of providing the parameters or the study object directly,
        to make it possible to create individual studies, when CustomOptuna optimize is called by an external wrapper
        (i.e. `cross_validate`).
        Further, the provided function is called with a single parameter `seed` that can be used to create a samplers
        and pruners with fixed random seeds.
        Using the passed seed compared to a fix global seed is important, as otherwise the same sampler/pruner use the
        same random numbers across multiple processes if `n_jobs > 1` (see more in Notes).
        Note, that this method should return consistent output when called multiple times with the same seed.
        Otherwise, unexpected behaviour can occur, where different processes use different samplers/pruners in a
        multi-processing setting.
    n_trials
        The number of trials.
        If this argument is set to `None`, there is no limitation on the number of trials.
        In this case you should use `timeout` instead.
        Because optuna is called internally by this wrapper, you can not set up a study without limits and end it
        using CTRL+C (as suggested by the Optuna docs).
        In this case the entire execution flow would be stopped.
    timeout
        Stop study after the given number of second(s).
        If this argument is set to `None`, the study is executed without time limitation.
        In this case you should use `n_trials` to limit the execution.
    return_optimized
        If True, a pipeline object with the overall best parameters is created and re-optimized using all provided data
        as input.
        The optimized pipeline object is stored as `optimized_pipeline_`.
        How the "re-optimization" works depends on the type of pipeline provided.
        If it is a simple pipeline, no specific re-optimization will be perfomed and `optimized_pipeline_` will simply
        be an instance of the pipeline with the best parameters indentified in the search.
        When `pipeline` is a subclass of `OptimizablePipeline`, we attempt to call `pipeline.self_optimize` with the
        entire dataset provided to the `optimize` method.
        The result of this self-optimization will be set as `optimized_pipeline`.
        If this behaviour is undesired, you can overwrite the `return_optimized_pipeline` method in subclass.s
    callbacks
        List of callback functions that are invoked at the end of each trial.
        Each function must accept two parameters with the following types in this order:
        :class:`~optuna.study.Study` and :class:`~optuna.trial.FrozenTrial`.
    n_jobs
        Number of parallel jobs to use (default = 1 -> single process, -1 -> all available cores).
        This uses joblib with the multiprocessing backend to parallelize the optimization.
        If this is set to -1, all available cores are used.

        .. warning:: Read the notes on multiprocessing below before using this feature.
    random_seed
        A random seed that is used as base for the seed passed to your implementation `get_study_params`.
        If None, this is set to a random integer between 0 and 100 (derived using numpy.random.randint).
        In case of multiprocessing, this seed is used as offset to create different seeds for each process.
    eval_str_paras
        This can be a sequence (tuple/list) of parameter names used by Optuna that should be evaluated using
        `literal_eval` instead of just set as string on the pipeline.
        The main usecase of this is to allow the user to pass a list of strings to `suggest_categorical` but have the
        actual pipeline recive the evaluated value of this string.
        This is required, as many storage backends of optuna only support number or strings as categorical values.

        A typical example would be wanting to select a set of axis for an algorithm that are expressed as a list/tuple
        of strings.
        In this case you would use a strinigfied version of these tuples as the categorical values in the optuna study
        and then use `eval_str_paras` to evaluate the stringified version to the actual tuple.

        >>> def search_space(trial):
        ...     trial.suggest_categorical("axis", ["('x',)", "('y',)", "('z',)", "('x', 'y')"])
        >>> optuna_opt = CustomOptunaOptimize(pipeline, ..., eval_str_paras=["axis"])

        Note, that in your custom subclass, you need to wrap the trial params in `self.sanitize_params` to make sure
        this sanitazation is applied to all parameters.

        >>> # Inside your custom subclass
        >>> def create_objective(self):
        ...     def objective(trial, pipeline, dataset):
        ...         params = self.sanitize_params(trial.params)
        ...         pipeline.set_params(**params)

        In all other places (e.g. when converting the final result table) this class handles sanitazation automatically.

    show_progress_bar
        Flag to show progress bars or not.
    gc_after_trial
        Run the garbage collector after each trial.
        Check the optuna documentation for more detail

    Other Parameters
    ----------------
    dataset
        The dataset instance passed to the optimize method

    Attributes
    ----------
    search_results_
        A dictionary containing all relevant results of the parameter search.
        The format of this dictionary is designed to be directly passed into the `pd.DataFrame` constructor.
        Each column then represents the result for one set of parameters.

        The dictionary contains the following entries:

        score
            The value of the score for a specific trial.
            If a trial was pruned this value is nan.
        param_{parameter_name}
            The value of a respective parameter.
        params
            A dictionary representing all parameters.
        state
            Whether the trial was completed, pruned, or any type of error occurred
        datetime_start
            When the trial was started
        datetime_complete
            When the trial ended
        duration
            The duration of the trial
        user_attrs_...
            User attributes set within the objective function
        system_attrs_...
            System attributes set internally by optuna (usually empty)

        If you need access to further parameter, inspect `self.study_` directly.
    optimized_pipeline_
        An instance of the input pipeline with the best parameter set.
        This is only available if `return_optimized` is not False.
    best_params_
        The parameter combination identified in the study
    best_score_
        The score achieved in the best trial
    best_trial_
        The trial object that resulted in the best results
    study_
        The study object itself.
        This should usually be identical to `self.study`.
    random_seed_
        The actual random seed used for the optimization.
        This is either the value passed to `random_seed` or a random integer between 0 and 100.

    Examples
    --------
    >>> from tpcp.validate import Scorer
    >>> from optuna import create_study
    >>> from optuna import samplers
    >>>
    >>> class MyOptunaOptimizer(CustomOptunaOptimize):
    ...     def create_objective(self):
    ...         def objective(trial: Trial, pipeline: Pipeline, dataset: Dataset):
    ...             trial.suggest_float("my_pipeline_para", 0, 3)
    ...             mean_score, _ = Scorer(lambda pipe, dp: pipe.score(dp))(pipeline, dataset)
    ...             return mean_score
    ...
    ...         return objective
    >>>
    >>> study = create_study(sampler=samplers.RandomSampler())
    >>> opti = MyOptunaOptimizer(pipeline=MyPipeline(), study=study, n_trials=10)
    >>> opti = opti.optimize(MyDataset())

    Notes
    -----
    Multiprocessing
    ***************
    This class provides a relatively hacky implementation of multiprocessing.
    The implementation is based on the suggestions made here: https://github.com/optuna/optuna/issues/2862
    However, it depends on internal optuna APIs and might break in future versions.

    To use multiprocessing, the provided `get_study_params` function must return a persistent backend (
    i.e. not the default InMemoryStorage), that can be written to by multiple processes.
    To make sure that your individual runs are independent and you don't leave behind any files, make sure you clean
    up your study after each run. You can use `optuna.delete_study(study_name=opti_instance.study_.study_name,
    storage=opti_instance.study_._storage)` for this.

    Further, if you use samplers, that use a random seed (you likely are), you need to make sure that you pass the
    `seed` provided to `get_study_params` to the sampler and potentially pruner as well.
    Otherwise, each sampler/pruner instance might use the same random seed, which will lead to identical trials in
    each process.

    From the implementation perspective, we split the number of trials into `n_jobs` chunks and then spawn one study
    per job.
    However, we first run the first trial in the main process, to make sure that the study is created correctly.
    Then we split the remaining trials into chunks and spawn a new process for each chunk.
    This study is a copy of the study from the main process and hence, points to the same database.
    Each process will then complete its chunk of trials and then terminate.
    This is a relatively naive implementation, but it avoids the overhead of spawning a new process for each trial.
    If this is always the best idea, is unclear.

    One downside of using multiprocessing is, that your runs will not be reproducible, as the order of the trials is not
    guraranteed and depends on when the individual processes finish.
    This can lead to different suggested parameters when non-trivial samplers are used.
    Note that this is not a specific problem of our implementation, but a general problem of using multiprocessing with
    optuna.

    Further, the use of `show_progress_bar` is not recommended when using multiprocessing, as one progress bar per
    process is created and the output is not very readable.
    It might still be helpful to see that something is happening.

    .. note:: Using SQLite as backend is known to cause issues with multiprocessing, when the database is
              stored on a network drive (e.g. as typically done on a cluster).
              On most clusters, you should use the local storage of your node for the database or use a
              different backend (e.g. Redis, MySQL), if multiple nodes need to access the database at once.

    """

    def __init__(
        self,
        pipeline: PipelineT,
        get_study_params: Callable[[int], StudyParamsDict],
        *,
        n_trials: Optional[int] = None,
        timeout: Optional[float] = None,
        callbacks: Optional[list[Callable[[Study, FrozenTrial], None]]] = None,
        gc_after_trial: bool = False,
        n_jobs: int = 1,
        random_seed: Optional[int] = None,
        eval_str_paras: Sequence[str] = (),
        show_progress_bar: bool = False,
        return_optimized: bool = True,
    ) -> None:
        self.pipeline = pipeline
        self.get_study_params = get_study_params

        self.n_trials = n_trials
        self.random_seed = random_seed
        self.timeout = timeout
        self.callbacks = callbacks
        self.gc_after_trial = gc_after_trial
        self.show_progress_bar = show_progress_bar
        self.n_jobs = n_jobs
        self.eval_str_paras = eval_str_paras
        self.return_optimized = return_optimized

    @staticmethod
    def as_dataclass():
        """Return a version of the Dataset class that can be subclassed using dataclasses."""
        import dataclasses  # pylint: disable=import-outside-toplevel

        @dataclasses.dataclass(eq=False, repr=False, order=False)
        class CustomOptunaOptimizeDc(_CustomOptunaOptimize[PipelineT, DatasetT]):
            """Dataclass version of CustomOptunaOptimize."""

            pipeline: PipelineT
            get_study_params: Callable[[int], StudyParamsDict]

            # Optuna Parameters that are directly forwarded to study.optimize
            n_trials: Optional[int] = None
            timeout: Optional[float] = None
            callbacks: Optional[list[Callable[[Study, FrozenTrial], None]]] = None
            gc_after_trial: bool = False
            show_progress_bar: bool = False

            n_jobs: int = 1
            random_seed: Optional[int] = None
            return_optimized: bool = True

            eval_str_paras: Sequence[str] = ()

            optimized_pipeline_: PipelineT = dataclasses.field(init=False, repr=False)
            study_: Study = dataclasses.field(init=False, repr=False)

        return CustomOptunaOptimizeDc

    @staticmethod
    def as_attrs():
        """Return a version of the Dataset class that can be subclassed using `attrs` defined classes.

        Note, this requires `attrs` to be installed!
        """
        from attrs import define, field  # pylint: disable=import-outside-toplevel

        @define(eq=False, repr=False, order=False, kw_only=True, slots=False)
        class CustomOptunaOptimizeAt(_CustomOptunaOptimize[PipelineT, DatasetT]):
            """Attrs version of CustomOptunaOptimize."""

            pipeline: PipelineT
            get_study_params: Callable[[int], StudyParamsDict]

            # Optuna Parameters that are directly forwarded to study.optimize
            n_trials: Optional[int] = None
            timeout: Optional[float] = None
            callbacks: Optional[list[Callable[[Study, FrozenTrial], None]]] = None
            gc_after_trial: bool = False
            show_progress_bar: bool = False

            n_jobs: int = 1
            random_seed: Optional[int] = None
            return_optimized: bool = True

            eval_str_paras: Sequence[str] = ()

            optimized_pipeline_: PipelineT = field(init=False, repr=False)
            study_: Study = field(init=False, repr=False)

        return CustomOptunaOptimizeAt


T = TypeVar("T")


class OptunaSearch(_CustomOptunaOptimize[PipelineT, DatasetT]):
    """GridSearch equivalent using Optuna.

    An opinionated parameter optimization for simple (i.e. non-optimizable) pipelines that can be used as a
    replacement to GridSearch.

    Parameters
    ----------
    pipeline
        A tpcp pipeline with some hyper-parameters that should be optimized.
        This should be a normal (i.e. non-optimizable pipeline) when using this class.
    get_study_params
        A callable that returns a dictionary with the parameters that should be used to create the study (i.e. passed
        to `optuna.create_study`).
        Creating the study is handled via a callable, instead of providing the parameters or the study object directly,
        to make it possible to create individual studies, when CustomOptuna optimize is called by an external wrapper
        (i.e. `cross_validate`).
        Further, the provided fucntion is called with a single parameter `seed` that can be used to create a samplers
        and pruners with different random seeds.
        This is important for multi-processing (see more in Notes).
        Note, that this method should return consistent output when called multiple times with the same seed.
        Otherwise, unexcepected behaviour can occur, where different processes use different samplers/pruners in a
        multi-processing setting (`n_jobs > 1`).

    create_search_space
        A callable that takes a :class:`~optuna.trial.Trial` object as input and calls `suggest_*` methods on it to
        define the search space.
    scoring
        A callable that can score a single data point given a pipeline.
        This function should return either a single score or a dictionary of scores.
        If scoring is `None` the default `score` method of the pipeline is used instead.

        Note that if scoring returns a dictionary, `score_name` must be set to the name of the score that should be used
        for ranking.
    score_name
        The name of the score that should be used for ranking in case the scoring function returns a dictionary of
        values.
    n_trials
        The number of trials.
        If this argument is set to `None`, there is no limitation on the number of trials.
        In this case you should use `timeout` instead.
        Because optuna is called internally by this wrapper, you can not set up a study without limits and end it
        using CTRL+C (as suggested by the Optuna docs).
        In this case the entire execution flow would be stopped.
    timeout
        Stop study after the given number of second(s).
        If this argument is set to `None`, the study is executed without time limitation.
        In this case you should use `n_trials` to limit the execution.
    return_optimized
        If True, a pipeline object with the overall best parameters is created.
        The optimized pipeline object is stored as `optimized_pipeline_`.
    callbacks
        List of callback functions that are invoked at the end of each trial.
        Each function must accept two parameters with the following types in this order:
        :class:`~optuna.study.Study` and :class:`~optuna.trial.FrozenTrial`.
    n_jobs
        Number of parallel jobs to use (default = 1 -> single process, -1 -> all available cores).
        This uses joblib with the multiprocessing backend to parallelize the optimization.
        If this is set to -1, all available cores are used.

        .. warning:: Read the notes in :class:`~tpcp.optimize.optuna.CustomOptunaOptimize` on multiprocessing below
                     before using this feature.
    random_seed
        A random seed that is used as base for the seed passed to your implementation `get_study_params`.
        If None, this is set to a random integer between 0 and 100 (derived using numpy.random.randint).
        In case of multiprocessing, this seed is used as offset to create different seeds for each process.
    eval_str_paras
        This can be a sequence (tuple/list) of parameter names used by Optuna that should be evaluated using
        `literal_eval` instead of just set as string on the pipeline.
        The main usecase of this is to allow the user to pass a list of strings to `suggest_categorical` but have the
        actual pipeline recive the evaluated value of this string.
        This is required, as many storage backends of optuna only support number or strings as categorical values.

        A typical example would be wanting to select a set of axis for an algorithm that are expressed as a list/tuple
        of strings.
        In this case you would use a strinigfied version of these tuples as the categorical values in the optuna study
        and then use `eval_str_paras` to evaluate the stringified version to the actual tuple.

        >>> def search_space(trial):
        ...     trial.suggest_categorical("axis", ["('x',)", "('y',)", "('z',)", "('x', 'y')"])
        >>> optuna_opt = OptunaSearch(pipeline, ..., eval_str_paras=["axis"])

    show_progress_bar
        Flag to show progress bars or not.
    gc_after_trial
        Run the garbage collector after each trial.
        Check the optuna documentation for more detail

    Other Parameters
    ----------------
    dataset
        The dataset instance passed to the optimize method

    Attributes
    ----------
    search_results_
        A dictionary containing all relevant results of the parameter search.
        The format of this dictionary is designed to be directly passed into the `pd.DataFrame` constructor.
        Each column then represents the result for one set of parameters.

        The dictionary contains the following entries:

        score / {scorer-name}
            The aggregated value of a score over all data-points.
            If a single score is used for scoring, then the generic name "score" is used.
            Otherwise, multiple columns with the name of the respective scorer exist
        param_*
            The value of a respective parameter
        params
            A dictionary representing all parameters
        state
            Whether the trial was completed, pruned, or any type of error occurred
        datetime_start
            When the trial was started
        datetime_complete
            When the trial ended
        duration
            The duration of the trial
        user_attrs_...
            User attributes set within the objective function
        system_attrs_...
            System attributes set internally by optuna (usually empty)

        If you need access to further parameter, inspect `self.study_` directly.
    optimized_pipeline_
        An instance of the input pipeline with the best parameter set.
        This is only available if `return_optimized` is not False.
    best_params_
        The parameter combination identified in the study
    best_score_
        The score achieved in the best trial
    best_trial_
        The trial object that resulted in the best results
    study_
        The study object itself.
        This should usually be identical to `self.study`.
    multimetric_
        If the scorer returned multiple scores
    random_seed_
        The actual random seed used for the optimization.
        This is either the value passed to `random_seed` or a random integer between 0 and 100.

    """

    create_search_space: Callable[[Trial], None]
    scoring: ScorerTypes[PipelineT, DatasetT, T]
    score_name: Optional[str]

    multimetric_: bool

    def __init__(
        self,
        pipeline: PipelineT,
        get_study_params: Callable[[int], StudyParamsDict],
        create_search_space: Callable[[Trial], None],
        *,
        scoring: ScorerTypes[PipelineT, DatasetT, T],
        score_name: Optional[str] = None,
        n_trials: Optional[int] = None,
        timeout: Optional[float] = None,
        callbacks: Optional[list[Callable[[Study, FrozenTrial], None]]] = None,
        gc_after_trial: bool = False,
        n_jobs: int = 1,
        random_seed: Optional[int] = None,
        eval_str_paras: Sequence[str] = (),
        show_progress_bar: bool = False,
        return_optimized: bool = True,
    ) -> None:
        self.create_search_space = create_search_space
        self.scoring = scoring
        self.score_name = score_name

        self.pipeline = pipeline
        self.get_study_params = get_study_params
        self.n_trials = n_trials
        self.timeout = timeout
        self.callbacks = callbacks
        self.gc_after_trial = gc_after_trial
        self.n_jobs = n_jobs
        self.random_seed = random_seed
        self.eval_str_paras = eval_str_paras
        self.show_progress_bar = show_progress_bar
        self.return_optimized = return_optimized

    def create_objective(self) -> Callable[[Trial, PipelineT, DatasetT], Union[float, Sequence[float]]]:
        """Create the objective function for optuna.

        This is an internal function and should not be called directly.
        """
        # Here we define our objective function

        scoring = _validate_scorer(self.scoring, self.pipeline)

        # In a multiprocessing case, this function is called once in the main process and then in the worker processes.
        # The first call in the main process allows us to set the multimetric_ attribute.
        def objective(trial: Trial, pipeline: PipelineT, dataset: DatasetT) -> float:
            # First we need to select parameters for the current trial
            if self.create_search_space is None:
                raise ValueError("No valid search space parameter.")
            self.create_search_space(trial)
            # Then we apply these parameters to the pipeline
            pipeline = pipeline.set_params(**self.sanitize_params(trial.params))

            average_scores, single_scores = scoring(pipeline, dataset)

            if not hasattr(self, "multimetric_"):
                if isinstance(average_scores, dict):
                    self.multimetric_ = True
                else:
                    self.multimetric_ = False

            if self.multimetric_:
                if self.score_name is None:
                    raise ValueError("score_name must be set if scoring returns a dictionary of scores")
                if self.score_name not in average_scores:
                    raise ValueError(f"score_name '{self.score_name}' not in scoring results ({average_scores.keys()})")
                score = average_scores[self.score_name]
            else:
                if self.score_name is not None:
                    warnings.warn(
                        "score_name is ignored if scoring returns a single score",
                        UserWarning,
                        stacklevel=2,
                    )
                score = average_scores

            # As a bonus, we use the custom params option of optuna to store the individual scores per datapoint and the
            # respective data labels
            if self.multimetric_:
                trial.set_user_attr("__average_scores", average_scores)
                trial.set_user_attr("__single_scores", single_scores)
            else:
                # No need to set average scores, as their is only one score and that is returned by the objective
                # function and will be included in the results by optuna
                trial.set_user_attr("__single_scores", {"score": single_scores})
            trial.set_user_attr("__data_labels", dataset.group_labels)

            return score

        return objective

    @property
    def search_results_(self) -> dict[str, Sequence[Any]]:  # noqa: D102
        search_results = super().search_results_
        search_results["data_labels"] = search_results.pop("user_attrs___data_labels")

        if self.multimetric_:
            search_results.pop("score")
            if average_scores := search_results.pop("user_attrs___average_scores", None):
                search_results.update(_invert_list_of_dicts(average_scores))

        if single_scores := search_results.pop("user_attrs___single_scores", None):
            search_results.update({f"single_{k}": v for k, v in _invert_list_of_dicts(single_scores).items()})

        # We add params back to the end of the dict to make it easier to read
        search_results["params"] = [self.sanitize_params(p) for p in search_results.pop("params")]
        return search_results

    def return_optimized_pipeline(
        self,
        pipeline: PipelineT,
        dataset: DatasetT,  # noqa: ARG002
        study: Study,
    ) -> PipelineT:
        """Return the pipeline with the best parameters of a study.

        This is an internal function and should not be called directly.
        """
        # Pipeline that will be passed here is already cloned, so no need to clone again.
        pipeline_with_best_params = pipeline.set_params(**self.sanitize_params(study.best_params))
        return pipeline_with_best_params


def _invert_list_of_dicts(list_of_dicts: list[dict[str, Any]]) -> dict[str, list]:
    inverted = defaultdict(list)
    for d in list_of_dicts:
        for k, v in d.items():
            inverted[k].append(v)
    return dict(inverted)



"""Utils to perform snapshot tests easily.

This is inspired by github.com/syrusakbary/snapshottest.
Note that it can not be used in combination with this module!
"""
import re
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal


class SnapshotNotFoundError(Exception):
    pass


class PyTestSnapshotTest:
    """Perform snapshot tests in pytest.

    This supports standard datatypes and scientific datatypes like numpy arrays and pandas DataFrames.

    To use this in your tests, add the following lines to your conftest.py:

    .. code-block:: python

        @pytest.fixture()
        def snapshot(request):
            with PyTestSnapshotTest(request) as snapshot_test:
                yield snapshot_test


        def pytest_addoption(parser):
            group = parser.getgroup("snapshottest")
            group.addoption(
                "--snapshot-update",
                action="store_true",
                default=False,
                dest="snapshot_update",
                help="Update the snapshots.",
            )

    This will register the snapshot fixture that you can use in your tests.
    Further, it will register the `--snapshot-update` commandline flag, which you can use to update the snapshots.

    To use the fixture in your tests, simply add it as a parameter to your test function:

    .. code-block:: python

        def test_my_test(snapshot):
            result = my_calculation()
            snapshot.assert_match(result, "my_result_1")

    This will store the result of `my_calculation()` in a snapshot file in a folder called `snapshot` in the same folder
    as the test file.
    The name of the snapshot file will be the name of the test function, suffixed with `_my_result_1`.
    When the test is run again, the result will be compared to the stored snapshot.

    To update a snapshot, either delete the snapshot file and manually run the test again or run pytest with the
    `--snapshot-update` flag.

    """

    curr_snapshot: str

    def __init__(self, request=None) -> None:
        self.request = request
        self.curr_snapshot_number = 0
        super().__init__()

    @property
    def _update(self):
        return self.request.config.option.snapshot_update

    @property
    def _module(self):
        return Path(self.request.node.fspath.strpath).parent

    @property
    def _snapshot_folder(self):
        return self._module / "snapshot"

    @property
    def _file_name_json(self):
        return self._snapshot_folder / f"{self._test_name}.json"

    @property
    def _file_name_csv(self):
        return self._snapshot_folder / f"{self._test_name}.csv"

    @property
    def _file_name_txt(self):
        return self._snapshot_folder / f"{self._test_name}.txt"

    @property
    def _test_name(self):
        cls_name = getattr(self.request.node.cls, "__name__", "")
        flattened_node_name = re.sub(r"\s+", " ", self.request.node.name.replace(r"\n", " "))
        return "{}{}_{}".format(f"{cls_name}." if cls_name else "", flattened_node_name, self.curr_snapshot)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _store(self, value):
        self._snapshot_folder.mkdir(parents=True, exist_ok=True)
        if isinstance(value, pd.DataFrame):
            value.to_json(self._file_name_json, indent=4, orient="table", date_unit="ns")
        elif isinstance(value, np.ndarray):
            np.savetxt(self._file_name_csv, value, delimiter=",")
        elif isinstance(value, str):
            with Path(self._file_name_txt).open("w") as f:
                f.write(value)
        else:
            raise TypeError(f"The dtype {type(value)} is not supported for snapshot testing")

    def _retrieve(self, dtype):
        if dtype == pd.DataFrame:
            filename = self._file_name_json
            if not filename.is_file():
                raise SnapshotNotFoundError
            return pd.read_json(filename, orient="table")
        if dtype == np.ndarray:
            filename = self._file_name_csv
            if not filename.is_file():
                raise SnapshotNotFoundError
            return np.genfromtxt(filename, delimiter=",")
        if dtype == str:
            filename = self._file_name_txt
            if not filename.is_file():
                raise SnapshotNotFoundError
            with Path(self._file_name_txt).open() as f:
                value = f.read()
            return value
        raise ValueError(f"The dtype {dtype} is not supported for snapshot testing")

    def assert_match(self, value: Union[str, pd.DataFrame, np.ndarray], name: Optional[str] = None, **kwargs):
        """Assert that the value matches the snapshot.

        This compares the value with a stored snapshot of the same name.
        If no snapshot exists, it will be created.

        The snapshot name is automatically generated from the test name and the `name` parameter passed to this
        function.
        If no name is passed, the name will be suffixed with a number, starting at 0.
        If you have multiple snapshots in one test, we highly recommend to pass a name to this function.
        Otherwise, changing the order of the snapshots will break your tests.

        Parameters
        ----------
        value
            The value to compare with the snapshot.
            We support strings, numpy arrays and pandas DataFrames.
            For other datatypes like floats or short lists, we recommend to just use the standard pytest assertions
            and hardcode the expected value.
        name
            Optional name suffix of the snapshot-file.
            If not provided the name will be suffixed with a number, starting at 0.
        kwargs
            Additional keyword arguments passed to the comparison function.
            This is only supported for DataFrames and numpy arrays.
            There they will be passed to `assert_frame_equal` and `assert_array_almost_equal` respectively.

        """
        self.curr_snapshot = name or str(self.curr_snapshot_number)
        if self._update:
            self._store(value)
        else:
            value_dtype = type(value)
            try:
                prev_snapshot = self._retrieve(value_dtype)
            except SnapshotNotFoundError:
                self._store(value)  # first time this test has been seen
            except:
                raise
            else:
                if isinstance(value, pd.DataFrame):
                    assert_frame_equal(value, prev_snapshot, **kwargs)
                elif isinstance(value, np.ndarray):
                    np.testing.assert_array_almost_equal(value, prev_snapshot, **kwargs)
                elif isinstance(value, str):
                    # Display the string diff line by line as part of error message using difflib
                    import difflib

                    diff = difflib.ndiff(value.splitlines(keepends=True), prev_snapshot.splitlines(keepends=True))
                    diff = "".join(diff)
                    assert value == prev_snapshot, diff
                else:
                    raise TypeError(f"The dtype {value_dtype} is not supported for snapshot testing")

        self.curr_snapshot_number += 1
